import toml
import os
import re
from .utils import is_env_format
from .exceptions import ConfigException


env_pattern = "ENV::(?P<name>[0-9a-zA-Z]+)::(?P<type>[a-zA-Z]+)(::(?P<default>.*))?$"


class Config(dict):

    def __init__(self, default=None):
        super().__init__(default or {})

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError("{} do not exist".format(e.args[0]))

    def __setattr__(self, key, value):
        self[key] = value

    def is_env(self, value):
        """
        Judge the value adhere the format
        :param value: the value from the Config dict
        :return: (Boolean, match object)
        """
        if isinstance(value, str):
            match = re.match(env_pattern, value)
            if match:
                is_env_format(match, PARAM_GENERATOR)
                return True, match
        return False, None

    def get_env(self, match):
        """
        Get value from system environment variable
        if non-existent, setting the value to default
        and convert to variable type given
        :param match: the object returned by function re.match
        :return: 
        """
        ret = os.environ.get(match.group('name'), match.group('default'))
        if ret:
            into = PARAM_GENERATOR[match.group('type').upper()]
            ret = into(ret)
            return ret

    def check_dict(self, value):
        """
        Setting the value, if it conform the format
        :param value: the value from the Config dict
        :return: 
        """
        if isinstance(value, dict):
            for each in value:
                is_environ, is_param = self.is_env(value[each])
                if is_environ:
                    value[each] = self.get_env(is_param)
                self.check_dict(value[each])

    def use(self, name, func):
        """
        register the function
        :param name: key of dict
        :param func: custom function for transforming type of data
        :return: 
        """
        if PARAM_GENERATOR.get(name.upper(), None):
            raise ConfigException("function {}:{} exist".format(name, func.__name__))
        PARAM_GENERATOR[name.upper()] = func

    def load(self, file):
        """
        Traversing the dict from the toml file
        Set all value that conform to the format
        :param file: 
        :return: 
        """
        with open(file, 'r') as cfile:
            config = toml.loads(cfile.read())
        self.check_dict(config)
        self.update(config)


def into_str(value):
    if isinstance(value, str):
        return value
    return str(value)


def into_int(value):
    if isinstance(value, int):
        return value
    return int(value)


def into_bool(value):
    if isinstance(value, str):
        value = value.lower()
        if value == "true":
            return True
        return False
    return False

PARAM_GENERATOR = {
    'STR': into_str,
    'INT': into_int,
    'BOOL': into_bool
}



import inspect
from nougat.exceptions import UnknownMiddlewareException
from nougat.exceptions import ConfigException


def is_middleware(func):
    """
    test whether it is a middleware
    :return: Boolean
    """
    args = list(inspect.signature(func).parameters.items())

    if not inspect.iscoroutinefunction(func):
        print("hello")
        raise UnknownMiddlewareException("middleware {} should be awaitable".format(func.__name__))

    if len(args) != 2:
        raise UnknownMiddlewareException("middleware {} should has 2 params named ctx and next".format(func.__name__))

    if args[0][0] != 'ctx':
        raise UnknownMiddlewareException("the first param's name of middleware {} should be ctx".format(func.__name__))

    if args[1][0] != 'next':
        raise UnknownMiddlewareException("the second param's name of middleware {} should be next".format(func.__name__))


def is_env_format(match, dict):
    """
    test whether it conform the standard
    :return: Boolean
    """
    if not match.group('name').isupper():
        raise ConfigException("<name>'{}' is not capitalized.".format(match.group('name')))

    if not match.group('type') in dict:
        raise ConfigException("<type>'{}' is out of list {}".format(match.group('type'), dict.keys()))


import httptools
import json
from urllib.parse import parse_qsl
from nougat.httpstatus import STATUS_CODES
from nougat.exceptions import HttpException


class Params(object):

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

    def __format_attr(self, attr_type, value):
        pass


class Context(object):

    def __init__(self, app, path, headers, ip, version, method):

        # base
        self.app = app

        self.path = path.decode()  # e.g. /hello?a=1

        path_parse = httptools.parse_url(path)  # TODO use yarl to parse url
        self.url = path_parse.path  # e.g. /hello
        self.query_string = path_parse.query  # e.g. a=1

        self.__version = version  # http version
        self.method = method

        # request
        self.headers = dict(headers)
        self.cookies = {}

        self.ip = self.__init_ip(ip)
        self.req_body = {}

        self.query = None

        self.json = None

        # params
        self.params = Params()

        # response
        self.res_header = {}
        self.type = "text/plain"
        self.res = None
        self.status = 200

        self.__init_query(self.query_string)
        self.__init__cookies()

    @property
    def content_type(self):
        """
        the content type for request
        """
        return self.headers.get('CONTENT_TYPE', '').lower()

    def set_cookies(self, name, value, expires=None, domain=None, path=None, secure=False, http_only=False, same_site=None):
        """
        
        :param name: cookies name
        :param value: cookies value
        :param expires: number of seconds until the cookie expires
        :param domain: specifies those hosts to which the cookie will be sent
        :param path: indicates a URL path that must exist in the requested resource before sending the Cookie header
        :param secure: a secure cookie will only be sent to the server when a request is made using SSL and the HTTPS protocol 
        :param http_only: 
        :param same_site: 
        :return: 
        """

        header_value = "{}={}".format(name, value)

        if expires:
            header_value = "{}; Max-Age={}".format(header_value, expires)
        if domain:
            header_value = "{}; Domain={}".format(header_value, domain)
        if path:
            header_value = "{}; Path={}".format(header_value, path)
        if secure:
            header_value = "{}; Secure".format(header_value)
        if http_only:
            header_value = "{}; HttpOnly".format(header_value)
        if same_site:
            header_value = "{}; SameSite={}".format(header_value, same_site)

        self.set("Set-Cookie", header_value)

    def set(self, key, value):
        """
        set response header
        :param key: the key of header
        :param value: the value of header
        """

        self.res_header[key] = value

    def url_for(self, name, **kwargs):
        """
        get the url according to the section name and handler name
        :param name: a string like section_name.handler_name
        :return: the url string
        """
        # TODO url for function
        _name_split = name.split(".")

        if len(_name_split) != 2:
            raise Exception()  # TODO new exception

        section_name, handler_name = _name_split
        section = self.app.sections.get(section_name, None)

        if not section:
            raise Exception()  # TODO ne exception

        route = section.get_route_by_name(handler_name)

        return route.url(**kwargs)

    def redirect(self, url):
        """
        redirect to another page
        :param url: the page need to go
        """
        self.set("Location", url)
        self.abort(302)

    def abort(self, status, body=None):
        """
        abort HTTPException
        :param status: http status code
        :param body: http body
        """
        raise HttpException(body, status)

    def __build_response_headers(self):
        """
        generator response headers
        :return: 
        """
        headers = []
        for key, value in self.res_header.items():
            headers.append("{}: {}\r\n".format(key, value).encode('utf-8'))

        return headers

    @property
    def output(self):
        """
        output the http payload
        """
        def body_bytes(body):
            body_type = type(body)
            if body_type is str:
                body = body.encode('utf-8')
            elif body_type is bytes:
                body = body
            else:
                body = b'Unable to interpret body'

            return body

        # TODO format the output
        body = body_bytes(self.res or STATUS_CODES.get(self.status, 'FAIL'))

        payload = []

        # HTTP STATUS
        payload.append('HTTP/{} {} {}\r\n'.format(self.__version, self.status, STATUS_CODES.get(self.status, 'FAIL')).encode('latin-1'))

        # CONTENT TYPE AND LENGTH
        payload.append('Content-Type: {}\r\n'.format(self.type).encode('latin-1'))
        payload.append('Content-Length: {}\r\n'.format(len(body)).encode('latin-1'))

        # HEADERS OF LOCATION OR COOKIES
        payload.extend(self.__build_response_headers())

        # RESPONSE BODY
        payload.append(b'\r\n')
        payload.append(body)

        return b''.join(payload)

    def __init_ip(self, ip):
        """
        set the right ip
        """
        if self.app.config.get("X-FORWARD-IP", False):
            return self.headers.get("X-Forwarded-For", "").strip().split(",")
        else:
            return ip

    def __init__cookies(self):
        """
        从 self.heanders 中读取 Cookies 并且格式化进入 self.cookies
        """
        cookies = self.headers.get("Cookie", None) or self.headers.get("cookie", None)
        if cookies:
            cookies = cookies.split("; ")
            for one_cookies in cookies:
                if one_cookies != "":
                    one_cookies = one_cookies.split("=")
                    self.cookies[one_cookies[0]] = "=".join(one_cookies[1:])

    def __init_query(self, query):
        """
        格式化 query 的内容
        :param query:
        :return:
        """
        if query:

            querys = query.decode("utf-8").split("&")
            for each_query in querys:
                query_path = each_query.split("=")
                if len(query_path) >= 2:
                    self.__set_query(query_path[0], "=".join(query_path[1:]))

    def __set_query(self, key, value):
        if key in self.query:
            if not isinstance(self.query[key], list):
                temp = list()
                temp.append(self.query[key])
                self.query[key] = temp
            self.query[key].append(value)

        else:
            self.query[key] = value

    def init_body(self, body):
        """
        format body into different type
        :param body:
        :return:
        """
        # parse body as json
        if self.content_type == 'application/json':
            self.json = json.loads(body.decode())
            return

        if not self.content_type.startswith('multipart/'):
            pairs = parse_qsl(body.decode())
            for key, value in pairs:
                self.__set_body(key, value)
            return

        # TODO: from-data and file

    def __set_body(self, key, value):

        if key in self.body:
            if not isinstance(self.body[key], list):
                temp = list()
                temp.append(self.body[key])
                self.body[key] = temp
            self.body[key].append(value)

        else:
            self.body[key] = value

    def __get_msg(self, from_where, key, append=False):
        ret = None
        if from_where == 'cookie':
            return self.cookies.get(key, None)

        elif from_where == "query":
            ret = self.query.get(key, None)

        elif from_where == "form":
            ret = self.body.get(key, None)

        elif from_where == "header":
            return self.headers.get(key, None)

        if isinstance(ret, list):
            return ret if append else ret[-1]

        return ret

    def generate_params(self, route):
        """
        格式化参数
        :param route:
        :return:
        """
        for param in route.params:

            if param.name in route.url_params_dict:
                self.params.__setattr__(param.name, route.url_params_dict[param.name])
                continue

            for location in param.location:
                # pprint("{} {} {}", location, param.name, param.append)
                param_name = param.action or param.name
                param_content = self.__get_msg(location, param.name, param.append) or param.default

                self.params.__setattr__(param_name, param_content)



import inspect
from functools import partial
from nougat.router import Router
from nougat.exceptions import HandlerRedefineException, NougatRuntimeError
from nougat.utils import is_middleware


__all__ = ['Section']


class Section:

    def __init__(self, name):
        self.name = name
        self.router = Router(self.name)
        self.__temper_params = []
        self.chain = []
        self.__handler = {}

    def __route(self, url, method):
        """
        添加路由
        :param url: 路由规则
        :param method: HTTP 访问方法
        """
        def response(handler):
            if handler.__name__ in self.__handler:
                raise HandlerRedefineException(self.name, handler.__name__)
            route = self.router.add(url, handler, self.name, method, self.__temper_params)
            self.__handler[handler.__name__] = route
            self.__temper_params = []  # reset temper params
            return handler
        return response

    def get_route_by_name(self, handler_name):
        """
        get route by its handler name
        """
        return self.__handler.get(handler_name, None)

    def head(self, url):
        return self.__route(url, "HEAD")

    def get(self, url):

        return self.__route(url, "GET")

    def post(self, url):

        return self.__route(url, "POST")

    def put(self, url):

        return self.__route(url, "PUT")

    def delete(self, url):

        return self.__route(url, "DELETE")

    def options(self, url):

        return self.__route(url, "OPTIONS")

    def patch(self, url):

        return self.__route(url, "PATCH")

    def param(self,  name, type, location='query', optional=False, default=None, action=None, append=False, description=None):
        """
        为路由添加参数
        :param name: 参数名称
        :param type:参数类型
        :param location: 参数来源
        :param optional: 参数是否可选
        :param default: 参数默认值，仅当 optional 为 True 时有效
        :param action: 参数重命名
        :param append: 参数是否为 list
        :param description: 参数描述
        """
        self.__temper_params.append(dict(name=name,
                                         type=type,
                                         location=location,
                                         optional=optional,
                                         default=default,
                                         action=action,
                                         append=append,
                                         description=description))

        def response(handler):

            return handler

        return response

    def use(self, middleware):
        """
        register Middleware
        """
        if inspect.isfunction(middleware):
                is_middleware(middleware)
                self.chain.insert(0, middleware)
        else:
            raise NougatRuntimeError("section only can use middleware function")

    async def handler(self, context, route):

        async def ret_handler(ctx, next):
            ret = await next()
            # TODO handle different type of ret: json, text, html
            ctx.res = ret

        handler = route.handler
        handler = partial(handler, ctx=context)
        handler = partial(ret_handler, ctx=context, next=handler)

        for middleware in self.chain:
            handler = partial(middleware, ctx=context, next=handler)

        await handler()



from nougat.exceptions import HttpException

__all__ = ['STATUS_CODES', 'abort']
STATUS_CODES = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',

    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    208: 'Already Reported',
    226: 'IM Used',

    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    308: 'Permanent Redirect',

    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    418: 'I\'m a teapot',
    420: 'Enhance Your Calm',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    425: 'Reserved for WebDAV',
    426: 'Upgrade Required',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    431: 'Request Header Fields Too Large',
    444: 'No Response',
    449: 'Retry With',
    450: 'Blocked by Windows Parental Controls',
    451: 'Unavailable For Legal Reasons',
    499: 'Client Closed Request',

    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    506: 'Variant Also Negotiates',
    507: 'Insufficient Storage',
    508: 'Loop Detected',
    509: 'Bandwidth Limit Exceeded',
    510: 'Not Extended',
    511: 'Network Authentication Required',
    598: 'Network read timeout error',
    599: 'Network connect timeout error',

}


def abort(status, body=None):
    if not body:
        body = STATUS_CODES.get(status, "User Definition Code")
    raise HttpException(body, status)




from nougat.app import *
from nougat.exceptions import *
from nougat.httpstatus import *
from nougat.protocol import *
from nougat.router import *
from nougat.section import *
from nougat.test_client import *

__version__ = "0.1.0"



from setuptools import setup
from os import path
from nougat import __version__
from distutils.errors import DistutilsPlatformError

here = path.abspath(path.dirname(__file__))


setup_kwargs = {
    'name': 'nougat',
    'version': __version__,
    'description': 'Async API framework with human friendly params definition and automatic document',
    'url': 'https://github.com/Kilerd/nougat',

    'author': 'Kilerd Chan',
    'author_email': 'blove694@gmail.com',

    'license': 'MIT',

    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    'keywords': 'web framework async',

    'packages': ['nougat'],
    'install_requires': [
        'httptools>=0.0.9',
        'uvloop>=0.8.0',
        'aiofiles>=0.3.0',
        'aiohttp>=2.0.7',
        'toml>=0.9.2'
    ],
}
try:
    setup(**setup_kwargs)

except DistutilsPlatformError as exception:

    setup_kwargs['install_requires'].remove('uvloop>=0.8.0')
    setup(**setup_kwargs)



class HttpException(Exception):

    def __init__(self, body, status):
        super().__init__()
        self.body = body
        self.status = status


class ConfigException(Exception):

    def __init__(self, err=None):
        if err:
            self.err = err
        else:
            self.err = 'Config is nonstandard'

    def __str__(self):
        return "Format : 'ENV::<name>::<type>::<default_value>'\n" + self.err


class UnknownMiddlewareException(Exception):

    def __init__(self, err=None):
        if err:
            self.err = err
        else:
            self.err = 'Unknown middleware exception'

    def __str__(self):
        return self.err


class UnknownSectionException(Exception): pass


class UnknownRouterException(Exception): pass


class RouteReDefineException(Exception):

    def __init__(self, method, url):
        self.method = method
        self.url = url

    def __str__(self):
        return "{} {} seems been redefined".format(self.method, self.url)


class NougatRuntimeError(Exception):

    def __init__(self, text=""):
        self.text = text


class HandlerRedefineException(Exception):

    def __init__(self, section, handler):
        self.section =section
        self.handler = handler

    def __str__(self):
        return "{} handler seems redefine in section {}".format(self.handler, self.section)


class ParamRedefineException(Exception):

    def __init__(self, rule, name):
        self.rule = rule
        self.name = name

    def __str__(self):
        return "{} seems redefine param named {}".format(self.rule, self.name)


class ParamMissingException(Exception):

    def __init__(self, rule, name):
        self.rule = rule
        self.name = name

    def __str__(self):
        return "{} seems miss param named {}".format(self.rule, self.name)
